/*-
 * Copyright (c) 2004 AIAN
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id:$
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "basil.h"
#include "bgb.h"

int usage(void);
int initialize(int *fd, const char * port, struct gbcart * cart);
int select_bank(int fd, unsigned short int bank);
int get_cart_info(int fd, struct gbcart * cart);
int print_nintendo_logo(int fd);
int print_cart_info(struct gbcart * cart);
int read_ROM_steadily(int fd, struct gbcart * cart);
int read_ROM_all(int fd, struct gbcart * cart);
static int create_file_name(char *file_name, struct gbcart * cart);
static char file_char_validation(char test_char);

int
main(int argc, char *argv[])
{
    int ch;
    int fd;
    size_t length;
    int hflag = 0;
    int iflag = 0;
    int lflag = 0;
    int pflag = 0;
    int rflag = 0;
    int sflag = 0;
    char *port = NULL;
    struct gbcart cart;

    if (argc < 2) {
        printf("bgb : option error.\n");
        usage();
        return -1;
    }
    while ((ch = getopt(argc, argv, "hilp:rS")) != -1)
    {
        switch (ch) {
        case 'h':
            hflag = 1;
            break;
        case 'i':
            iflag = 1;
            break;
        case 'l':
            lflag = 1;
            break;
        case 'p':
            length = strlen(optarg) + 1;
            port = (char *)malloc(sizeof(char) * length);
            if (port == NULL) {
                fprintf(stderr, "bgb : memory allocation error.\n");
            }
            strlcpy(port, optarg, length);
            pflag = 1;
            break;
        case 'r':
            rflag = 1;
            break;
        case 'S':
            sflag = 1;
            break;
        }
    }
    argc -= optind;
    argv += optind;
    
    /* set BASIL port name default */
    if (!pflag) {
        length = strlen(COMPORT) + 1;
        port = (char *)malloc(sizeof(char) * length);
        if (port == NULL) {
            fprintf(stderr, "bgb : memory allocation error.\n");
        }
        strlcpy(port, COMPORT, length);
    }
    /* processing */
    if (hflag) {
        iflag = lflag = pflag = rflag = sflag = 0;
        usage();
    } else {
        initialize(&fd, port, &cart);
    }
    
    if (iflag) {
        print_cart_info(&cart);
    } else if (lflag) {
        print_nintendo_logo(fd);
    } else if (rflag) {
        read_ROM_all(fd, &cart);
    } else if (sflag) {
        read_ROM_steadily(fd, &cart);
    }

    /* end */
    if (!hflag) {
        basil_close(fd);
    }
    free(port);
    
    return 0;
}

int 
usage()
{
    printf("syntax : bgb [[-p PORT] -i | -r | -l | -S] [-h]\n");
    printf("options:\n"
           "  -i       show ROM info.\n"
           "  -p port  specify basil port. (default '/dev/basil')\n"
           "  -r       read ROM fast.\n"
           "  -S       read ROM slowly but steady. [obsolete]\n"
           "  -l       show Nintendo logo for circuit's testing\n"
           "  -h       show this message.\n");

    return 0;
}

int 
initialize(int * fd, const char * port, struct gbcart * cart)
{
    *fd = basil_open(port);
    basil_reset(*fd);
    get_cart_info(*fd, cart);

    return 0;
}

int
ROM_bank_select(int fd, struct gbcart * cart, unsigned short int bank)
{
    unsigned char bank_low = 0;
    unsigned char bank_high = 0;

    switch (CARTTYPE[cart->type]) {
    case ROMONLY:
        break;
    case MBC1:
        bank_high = (unsigned char) (bank & 0xE0) >> 5;
        bank_low = (unsigned char) bank & 0x1F;
        write_rom(fd, 0x2000, bank_low);
        write_rom(fd, 0x4000, bank_high);
        break;
    case MBC2:
        bank_low = (unsigned char) (bank & 0xFF);
        write_rom(fd, 0x2100, bank);
        break;
    case MBC3:
        bank_low = (unsigned char) (bank & 0xFF);
        write_rom(fd, 0x2000, bank);
        break;
    case MBC5:
        bank_high = (unsigned char) (bank & 0xFF00) >> 8;
        bank_low = (unsigned char) bank & 0x00FF;
        write_rom(fd, 0x2000, bank_low);
        write_rom(fd, 0x3000, bank_high);
        break;
    case HUC1: /* N/A now*/
        bank_high = (unsigned char) (bank & 0xE0) >> 5;
        bank_low = (unsigned char) bank & 0x1F;
        write_rom(fd, 0x2000, bank_low);
        write_rom(fd, 0x4000, bank_high);
        break;
    default:
        return -1;  /* NOT yet supported Camera, HuC-3, etc ... */
    }
    
    return 0;
}

int
get_cart_info(int fd, struct gbcart * cart)
{
    unsigned char c;
    int i;

    for (i = 0; i < TITLE_MAX; i++) {
	cart->title[i] = read_rom(fd, 0x0134 + i);
    }
    cart->title[TITLE_MAX] = '\0';
    cart->type = read_rom(fd, 0x0147);
    c = read_rom(fd, 0x0148);
    cart->rom_size = ROMSIZE[c];
    cart->rom_bank = ROMBANK[c];
    c = read_rom(fd, 0x0149);
    cart->ram_size = RAMSIZE[c];
    cart->ram_bank = RAMBANK[c];
    cart->checksum = read_rom(fd, 0x014E) << 8;
    cart->checksum += read_rom(fd, 0x014F);

    return 0;
}

/* dump Nintendo logo - just a test func. */
int
print_nintendo_logo(int fd)
{
    int i;
    int j = 0;
    unsigned char c;

    for (i = 0; i < 16; i++) {
	printf("+%X ", i);
    }
    printf("\n");
    for (i = 0x0104; i <= 0x0133; i++) {
	c = read_rom(fd, i);
	printf("%02x ", c);
	if (j == 15) {
	    printf("\n");
	    j = 0;
	} else {
	    j++;
	}
    }
    printf("\n");

    return 0;
}

int
print_cart_info(struct gbcart * cart)
{
    printf("----------------------------------\n");
    printf("TITLE   : %s\n", cart->title);
    printf("CARTTYPE: [%02X] %s\n", cart->type,CARTTYPE_LABEL[cart->type]);
    printf("ROM SIZE: %d byte%c (%d bank%c)\n",
           cart->rom_size, (cart->rom_size < 2)?'\0':'s',
           cart->rom_bank, (cart->rom_bank < 2)?'\0':'s');
    printf("RAM SIZE: %d byte%c (%d bank%c)\n",
           cart->ram_size, (cart->ram_size < 2)?'\0':'s',
           cart->ram_bank, (cart->ram_bank < 2)?'\0':'s');
    printf("CHECKSUM: %X\n", cart->checksum);
    printf("----------------------------------\n");

    return 0;
}

int
read_ROM_steadily(int fd, struct gbcart * cart)
{
    int i;
    char file_name[FILE_NAME_MAX] = "";
    unsigned short int bank;
    unsigned char data[BANK_BYTE];
    FILE *fp;

    create_file_name(file_name, cart);
    if ((fp = fopen(file_name, "w")) == NULL) {
	fprintf(stderr, "file open error!\n");
    }
    printf(">>>> READING ROM\n");
    printf(">>>> [file name] %s\n", file_name);
    print_cart_info(cart);
    for (bank = 0; bank < cart->rom_bank; bank++) {
	printf("\rREADING BANK %03d/%03d ...", bank, cart->rom_bank - 1);
	fflush(NULL);
	if (bank == 0) {
	    address_set(fd, 0x0000);
	    for (i = 0x0000; i < 0x4000; i++) {
		data[i] = read_rom_with_address_increment(fd);
	    }
	} else {
	    ROM_bank_select(fd, cart, bank);
	    address_set(fd, 0x4000);
	    for (i = 0x0000; i < 0x4000; i++) {
		data[i] = read_rom_with_address_increment(fd);
	    }
	}
	fwrite(data, sizeof(unsigned char), BANK_BYTE, fp);
    }
    printf("\nDONE.\n\a");

    return 0;
}

int
read_ROM_all(int fd, struct gbcart * cart)
{
    unsigned short int bank;
    char file_name[FILE_NAME_MAX] = "";
    unsigned char block[BLOCK_BYTES];
    FILE *fp;
    int i;

    create_file_name(file_name, cart);
    if ((fp = fopen(file_name, "w")) == NULL) {
	fprintf(stderr, "file open error!\n");
    }
    printf(">>>> READING ROM\n");
    printf(">>>> [file name] %s\n", file_name);
    print_cart_info(cart);

    for (bank = 0; bank < cart->rom_bank; bank++) {
	printf("\rREADING BANK %03d/%03d ...", bank, cart->rom_bank - 1);
	fflush(NULL);
	if (bank == 0) {
            address_set(fd, 0x0000);
            for (i = 0; i < 64; i++) {
                read_rom_block(fd, block);
                fwrite(block, sizeof(unsigned char), BLOCK_BYTES, fp);
            }
        } else {
            ROM_bank_select(fd, cart, bank);
            address_set(fd, 0x4000);
            for (i = 0; i < 64; i++) {
                read_rom_block(fd, block);
                fwrite(block, sizeof(unsigned char), BLOCK_BYTES, fp);
            }
        }
    }
    printf("\nDONE.\n\a");
    
    return 0;
}

static int
create_file_name(char *file_name, struct gbcart * cart)
{
    char base_name[TITLE_MAX];
    int i;

    for (i = 0; i < TITLE_MAX; i++) {
        if (cart->title[i] == '\0') {
	    base_name[i] = '\0';
            i = TITLE_MAX;
            break;
        } else {
            base_name[i] = file_char_validation(cart->title[i]);
        }
    }
    /* strlcpy(3) and strlcat(3) are not in standard C library */
    strlcpy(file_name, base_name, TITLE_MAX);
    strlcat(file_name, ".gb", FILE_NAME_MAX);

    return 0;
}

static char
file_char_validation(char test_char)
{
    char valid_char;

    if (isalnum(test_char)) {
        valid_char = test_char;
    } else {
        valid_char = '_';
    }

    return valid_char;
}
